unit hiZipperEx;

interface

uses Kol,Windows,Share,Debug,Lib;

type
  TProcThunk = packed record
    POPEDX:Byte;
    MOVEAX:Byte;
    SelfPtr:Pointer;
    PUSHEAX:Byte;
    PUSHEDX:Byte;
    JMP:Byte;
    JmpOffset:Integer;
  end;

type
  TClock = packed record
    sec,min,hrs:byte;
  end;

type TOpertionType = (EndPack, EndUnpack, EndDelete, EndToBreak, EndToError);

type
 THIZipperEx = class( TDebug )
   private
    ProcThunk: TProcThunk;
    FTimer: PTimer;
    fRealTime: integer;
    fInfoRealTime: TClock;
    ftimetmp: integer;
    fBitSec: Int64;
    old_bs,new_bs: Int64;
    fAllTime: string;
    fAverageValue: integer;
    dt: TData;

    List                    : PStrList;
    ArrList                 : PArray;
    function _GetList(
                   var Item : TData;
                    var Val : TData
                      )     : boolean;
    function _CountList     : integer;
   public
    _prop_ZipFileName       : string;
    _prop_BasePath          : string;
    _prop_FileMask          : string;
    _prop_Password          : string;
    _prop_Comment           : string;
    _prop_SpanSize          : integer;
    _prop_UpdateMode        : byte;
    _prop_Method            : byte;
    _prop_SkipOlder         : byte;
    _prop_TestOnly          : byte;
    _prop_ResetAttr         : byte;
    _prop_CheckAttr         : byte;
    _prop_UseFolders        : byte;
    _prop_IncludeSubfolders : byte;
    _prop_OverwriteExisting : byte;
    _event_onProgress       : THI_Event;
    _event_onProgressMax    : THI_Event;
    _event_onInfo           : THI_Event;
    _event_onError          : THI_Event;
    _data_Data              : THI_Event;
    _data_ZipFileName       : THI_Event;
    _data_BasePath          : THI_Event;
    _data_FileMask          : THI_Event;
    _data_Index             : THI_Event;
    sysDateTime             : TFileTime;
    strZIP                  : string;
    strPath                 : string;
    strName                 : string;
    strMask                 : string;
    strStoredName           : string;
    intCount                : integer;
    intMethod               : integer;
    intMode                 : integer;
    fResetAttr              : boolean;
    fCheckAttr              : boolean;
    fOverwriteExisting      : boolean;
    fSkipOlder              : boolean;
    fIncludeSubfolders      : boolean;
    fUseFolders             : boolean;
    fTestOnly               : boolean;
    fBytes                  : integer;
    fFiles                  : integer;
    constructor Create;
    destructor Destroy; override;
    procedure _work_doCreate(var _Data: TData; Index: word);
    procedure _work_doOpen(var _Data: TData; Index: word);
    procedure _work_doOrder(var _Data: TData; Index: word);
    procedure _work_doOrderMulti(var _Data: TData; Index: word);
    procedure _work_doCompress(var _Data: TData; Index: word);
    procedure _work_doClose(var _Data: TData; Index: word);
    procedure _work_doExtractAll(var _Data: TData; Index: word);
    procedure _work_doExtractOne(var _Data: TData; Index: word);
    procedure _work_doList(var _Data: TData; Index: word);
    procedure _work_doDelete(var _Data: TData; Index: word);
    procedure _work_doCancel(var _Data: TData; Index: word);
    procedure _work_doPassword(var _Data: TData; Index: word);
    procedure _var_Count(var _Data: TData; Index: word);
    procedure _var_List(var _Data:TData; Index:word);
    procedure _var_TotalBytes(var _Data:TData; Index:word);
    procedure _var_TotalSizeStr(var _Data:TData; Index:word);
    procedure _var_CompressionRatio(var _Data:TData; Index:word);
    function  Check(intResult: integer) : boolean;
    procedure onProgress(); stdcall;
    procedure OnTimer(Obj:PObj);
    procedure CalcEndTime(var itime:integer; var stime: string);
    procedure InfoTime(var time: string);
    procedure EndOperation(OperationType: TOpertionType);

  end;

const
  OLE32 = 'OLE32.dll';
  AVERAGE_TIME = 3;

implementation

uses HiTime;

function BytesToStr(const i64Size: Int64): string;
const
  i64GB = (1024 * 1024) * 1024;
  i64MB = 1024 * 1024;
  i64KB = 1024;
begin
  if i64Size div i64GB > 0 then
    Result := Double2str(round(i64Size/i64GB/0.01)*0.01)+' Gb'
  else if i64Size div i64MB > 0 then
    Result := Double2str(round(i64Size/i64MB/0.01)*0.01)+' Mb'
  else if i64Size div i64KB > 0 then
    Result := Double2str(round(i64Size/i64KB/0.01)*0.01)+' Kb'
  else
    Result := Int2Str(i64Size) + ' Byte';
end;

function CoDosDateTimeToFileTime(
  nDosDate              : Word;
  nDosTime              : Word;
  var filetime          : TFileTime
  ) : BOOLEAN; stdcall;
  external OLE32 name 'CoDosDateTimeToFileTime';


constructor THIZipperEx.Create;
begin
  inherited Create;
  List := NewStrList;
  ArrList := CreateArray( nil, _GetList, _CountList, nil );
  strZIP := '';
  ProcThunk.POPEDX    := $5A;
  ProcThunk.MOVEAX    := $B8;
  ProcThunk.SelfPtr   := Self;
  ProcThunk.PUSHEAX   := $50;
  ProcThunk.PUSHEDX   := $52;
  ProcThunk.JMP       := $E9;
  ProcThunk.JmpOffset := Integer(@THIZipperEx.onProgress) -
                        Integer(@ProcThunk.JMP) - 5;

  FTimer := NewTimer(1000);
  FTimer.Enabled := False;
  FTimer.OnTimer := onTimer;
  fRealTime := 0;
  dtNull(dt);
end;

destructor THIZipperEx.Destroy;
begin
  List.Free;
  if ArrList <> nil then dispose(ArrList);
{$ifndef F_P}
  FTimer.Free;
{$endif}
  inherited;
end;

procedure THIZipperEx.EndOperation(OperationType: TOpertionType);
begin
  FTimer.Enabled := False;
  _hi_OnEvent( _event_onProgressMax, 0);
  _hi_OnEvent( _event_onProgress, 0);
  _hi_onEvent( _event_onInfo,' ; ; ; ; ; ');
end;

function FormatInd(i:integer): string;
begin
  case i of
   0..9: begin Result := '0'+ int2str(i); end;
  else  Result := int2str(i);
  end;
end;

procedure THIZipperEx.CalcEndTime(var itime:integer; var stime: string);
const iMIN = 60; iHRS = (60 * 60);
var s,m,h:integer;
begin
  h := Round(itime / iHRS);
  m := Round(itime / iMIN) mod 60;
  s := Round(itime) mod 60;
  stime := FormatInd(h) + ':' +
           FormatInd(m) + ':' +
           FormatInd(s);
end;

procedure THIZipperEx.InfoTime(var time: string);
begin
  inc(fRealTime);
  inc (fInfoRealTime.sec);
  case fInfoRealTime.sec of
    60: begin fInfoRealTime.sec := 0; inc(fInfoRealTime.min) end;
  end;
  case fInfoRealTime.min of
    60: begin fInfoRealTime.min := 0; inc(fInfoRealTime.hrs) end;
  end;
  case fInfoRealTime.hrs of
    24: fInfoRealTime.hrs := 0;
  end;

  fAverageValue := fAverageValue + ftimetmp;
  if ( fRealTime mod AVERAGE_TIME ) = 0 then
  begin
    ftimetmp := Trunc(fAverageValue / AVERAGE_TIME);
    CalcEndTime(ftimetmp,fAllTime);
    ftimetmp := 0; fAverageValue := 0;
  end;

  time := fAllTime + ';' + FormatInd(fInfoRealTime.hrs) + ':' +
                           FormatInd(fInfoRealTime.min) + ':' +
                           FormatInd(fInfoRealTime.sec) + ';';
end;

procedure THIZipperEx.OnTimer;
var
  TimeFmt, tmp: string;
  intFiles,intBytes: integer;
begin
  zGetRunTimeInfo( intFiles, intBytes );
  new_bs := intBytes;
  fBitSec := (new_bs - old_bs);
  InfoTime(TimeFmt);
  tmp := Int2Str(fFiles) + ';' + Int2Str(intFiles) + ';' +
  TimeFmt + BytesToStr(fBitSec) +  '/sec';
  _hi_onEvent(_event_onInfo, tmp );
  old_bs := new_bs;
  ftimetmp := Trunc(fBytes / fBitSec);
end;

procedure THIZipperEx._work_doCompress;
begin
  if strZIP = '' then Exit;
  old_bs := 0; new_bs := 0;;
  fBytes :=  zGetOrderedBytes();
  fFiles :=  zGetOrderedFiles();
  new_bs := fBytes;
  ftimetmp := 0; fAllTime := '00:00:00';
  FTimer.Enabled := True;
  _hi_OnEvent( _event_onProgressMax, fBytes );
  intMethod := _prop_Method;
  if _prop_ResetAttr = 0 then fResetAttr := True
  else fResetAttr := False;
  Check( zCompressFiles(
                   nil,
                   PChar( _prop_Password ),
                   intMethod,
                   fResetAttr,
                   _prop_SpanSize,
                   _prop_Comment,
                   @ProcThunk
                   ) );
  EndOperation(EndPack);
end;

function THIZipperEx.Check( intResult: integer ) : boolean;
begin
  Result := False;
  _hi_OnEvent( _event_onError, intResult );
  if intResult = 0 then Result := True
   else
     begin
       _work_doCancel(dt,0);
       Beep(900,200);
       ShowMessage(zGetLastErrorAsText());
     end;
end;

procedure THIZipperEx._work_doCreate;
begin
  if strZIP <> '' then
    if Check( zCloseZipFile ) then
      strZIP := '';
  strZIP := ReadString( _Data, _data_ZipFileName, _prop_ZipFileName );
  if Check( zCreateNewZip( PChar( strZIP ) ) ) then
    List.Clear
  else
    strZIP := '';
end;

procedure THIZipperEx._work_doOpen;
begin
  if strZIP <> '' then
    if Check( zCloseZipFile ) then
      strZIP := '';
  strZIP := ReadString( _Data, _data_ZipFileName, _prop_ZipFileName );
  if Check( zOpenZipFile( PChar( strZIP ) ) ) then
    List.Clear
  else
    strZIP := '';
end;

procedure THIZipperEx._work_doList;
const
  tab : string = Chr($09);
var
  i : Word;
  strTemp : string;
  date : word;
  time : word;
  sys : TSystemTime;
  Ratio : integer;
  fs : integer;
  cfs : integer;
begin
  if strZIP = '' then Exit;
  intCount := zGetTotalFiles;
  List.Clear;
  for i := 0 to intCount - 1 do begin
    strTemp := zGetFileName( i ) + tab;
    date := zGetFileDate( i );
    time := zGetFileTime( i );
    CoDosDateTimeToFileTime( date, time, sysDateTime );
    FileTimeToSystemTime( sysDateTime, sys );
    strTemp := strTemp + TimeToStr( 'D.M.Y h:m:s', sys ) + tab;
    fs := zGetFileSize( i );
    cfs := zGetCompressedFileSize( i );
    if fs <> 0 then
      Ratio := 100 - Trunc( ( cfs / fs ) * 100 )
    else
      Ratio := 0;
    strTemp := strTemp + BytesToStr( fs ) + tab;
    strTemp := strTemp + Int2Str( Ratio ) + '%' + tab;
    strTemp := strTemp + BytesToStr( cfs ) + tab;
    strTemp := strTemp + zGetFilePath( i );
    List.Add( strTemp );
  end;
end;

procedure THIZipperEx._work_doOrder;
begin
  if strZIP = '' then Exit;
  strName := ReadString( _Data, _data_Data, '' );
  strStoredName := kol.ExtractFileName( strName );
  intMode := _prop_UpdateMode;
  if Check( zOrderFile(
                   PChar( strName ),
                   PChar( strStoredName ),
                   intMode
                   ) ) then begin
    List.Add( kol.ExtractFileName( strName ) );
    intCount := List.Count;
  end;
end;

procedure THIZipperEx._work_doOrderMulti;
var
  idx: integer;
  strBasePath: string;
begin
  if strZIP = '' then Exit;
  strMask := ReadString( _Data, _data_FileMask, _prop_FileMask );
  strPath := ReadString( _Data, _data_BasePath, _prop_BasePath );
  if Copy( strPath, Length( strPath ), 1) <> '\' then strPath := strPath + '\';
  intMode := _prop_UpdateMode;
  idx := Length(strPath)-1;
  while (strPath[idx] <> '\') and (Length(strPath)> 0) do dec(idx);
  strBasePath := Copy( strPath, 1, idx);
  if _prop_IncludeSubfolders = 0 then fIncludeSubfolders := True
  else fIncludeSubfolders := False;
  if _prop_CheckAttr = 0 then fCheckAttr := True
  else fCheckAttr := False;
  Check( zOrderByWildcards(
                   PChar( strPath + strMask ),
                   PChar( strBasePath ),
                   fIncludeSubfolders,
                   fCheckAttr,
                   intMode
                   ) );
end;

procedure THIZipperEx._work_doExtractAll;
begin
  if strZIP = '' then Exit;
  strPath := ReadString( _Data, _data_BasePath, _prop_BasePath );
  if _prop_OverwriteExisting = 0 then fOverwriteExisting := True
  else fOverwriteExisting := False;
  if _prop_SkipOlder = 0 then fSkipOlder := True
  else fSkipOlder := False;
  if _prop_UseFolders = 0 then fUseFolders := True
  else fUseFolders := False;
  if _prop_TestOnly = 0 then fTestOnly := True
  else fTestOnly := False;

  old_bs := 0; new_bs := 0;;
  fBytes :=  zGetTotalBytes();
  fFiles :=  zGetTotalFiles();
  new_bs := fBytes;
  ftimetmp := 0; fAllTime := '00:00:00';
  FTimer.Enabled := True;
  _hi_OnEvent( _event_onProgressMax, fBytes );

  Check( zExtractAll(
                   PChar( strPath ),
                   PChar( _prop_Password ),
                   fOverwriteExisting,
                   fSkipOlder,
                   fUseFolders,
                   fTestOnly,
                   @ProcThunk
                   ) );
  EndOperation(EndUnpack);
end;

procedure THIZipperEx._work_doExtractOne;
var
  idx: integer;
begin
  if strZIP = '' then Exit;
  strPath := ReadString( _Data, _data_BasePath, _prop_BasePath );
  idx := ReadInteger( _Data, _data_Index, 0 );

  zSelectFile(idx, True);
  old_bs := 0; new_bs := 0;
  fBytes :=  zGetSelectedBytes();
  fFiles :=  zGetSelectedFiles();
  new_bs := fBytes;
  ftimetmp := 0; fAllTime := '00:00:00';
  FTimer.Enabled := True;
  _hi_OnEvent( _event_onProgressMax, fBytes );

  if _prop_OverwriteExisting = 0 then fOverwriteExisting := True
  else fOverwriteExisting := False;
  if _prop_SkipOlder = 0 then fSkipOlder := True
  else fSkipOlder := False;
  if _prop_UseFolders = 0 then fUseFolders := True
  else fUseFolders := False;
  if _prop_TestOnly = 0 then fTestOnly := True
  else fTestOnly := False;
  Check( zExtractOne(
                   idx,
                   PChar( strPath ),
                   PChar( _prop_Password ),
                   fOverwriteExisting,
                   fSkipOlder,
                   fUseFolders,
                   fTestOnly,
                   @ProcThunk
                   ) );
  EndOperation(EndUnpack);
end;

procedure THIZipperEx._work_doDelete;
var
  idx: integer;
begin
  if strZIP = '' then Exit;
  idx := ToInteger( _Data );
  if zSelectFile( idx, True) then begin
    Check( zDeleteFiles );
    zSelectFile( idx, False )
  end;
  EndOperation(EndDelete);
end;

procedure THIZipperEx._work_doCancel;
begin
  if strZIP = '' then Exit;
  if zCancelOperation = True then
  begin
    _hi_OnEvent( _event_onProgressMax, 0);
    _hi_OnEvent( _event_onProgress, 0);
    _hi_onEvent( _event_onInfo,' ; ; ; ; ; ');
    EndOperation(EndToBreak);
  end;
  FTimer.Enabled := False;
end;

procedure THIZipperEx._work_doClose;
begin
  if strZIP = '' then Exit;
  if Check( zCloseZipFile ) then
  begin
    List.Clear;
    strZIP := '';
    _hi_OnEvent( _event_onProgressMax, 0);
    _hi_OnEvent( _event_onProgress, 0);
    _hi_onEvent( _event_onInfo,' ; ; ; ; ; ');
    FTimer.Enabled := False;
  end;
end;

procedure THIZipperEx._work_doPassword;
begin
  _prop_Password := ToString( _Data );
end;

procedure  THIZipperEx._var_Count;
begin
  dtInteger(_Data, intCount);
end;

procedure THIZipperEx._var_List;
begin
  dtArray(_Data, ArrList );
end;

function THIZipperEx._GetList;
var
  ind: integer;
begin
  ind := ToIntIndex( Item );
  if (ind >= 0) and (ind < List.Count) then
   begin
    Result := true;
    dtString(Val,List.Items[ind]);
   end
  else Result := false;
end;

function THIZipperEx._CountList;
begin
  Result := List.Count;
end;

procedure THIZipperEx.onProgress;
var
  intFiles,intBytes: integer;
begin
  zGetRunTimeInfo( intFiles, intBytes );
  _hi_OnEvent( _event_onProgress,  intBytes );
end;

procedure THIZipperEx._var_TotalBytes;
var
  intTmp: integer;
begin
  intTmp := zGetTotalBytes();
  dtInteger(_Data,intTmp);
end;

procedure THIZipperEx._var_TotalSizeStr;
var
  str: string;
begin
  str := BytesToStr(zGetTotalBytes());
  dtString(_Data, str);
end;

procedure THIZipperEx._var_CompressionRatio;
var
  intTmp, intCompr: integer;
begin
  if strZIP = '' then Exit;
  intTmp := Trunc((zGetTotalBytes() / 100));
  intCompr := Trunc((FileSize(strZIP) / intTmp));
  dtInteger(_Data, intCompr);
end;

end.
